#[cfg(not(target_env = "msvc"))]
use tikv_jemallocator::Jemalloc;

#[cfg(not(target_env = "msvc"))]
#[global_allocator]
static GLOBAL: Jemalloc = Jemalloc;

use anyhow::{bail, Result};
use chrono::prelude::*;
use clap::{Parser, Subcommand};
use diesel::prelude::*;
use r2d2_redis::redis;
use scan_dir::ScanDir;
use serde::{Deserialize, Serialize};
use std::fs;
use std::path::PathBuf;

use backend::*;

#[derive(Debug, Parser)]
#[clap(author, version, about, long_about = None)]
struct Cli {
    #[clap(subcommand)]
    commands: Commands,
}

#[derive(Debug, Subcommand)]
enum Commands {
    #[clap(about = "Imports new posts")]
    Import,

    #[clap(about = "Clears redis cache")]
    Clear,
}

#[derive(Deserialize, Serialize, Debug)]
struct PostFrontmatter {
    id: Option<i32>,
    name: String,
    slug: String,
    description: String,
    published_at: NaiveDate,
    edited_at: Option<NaiveDate>,
    active: bool,
}

#[derive(Debug)]
struct Post {
    path: PathBuf,
    frontmatter: PostFrontmatter,
    content: String,
}

fn main() -> Result<()> {
    match Cli::parse().commands {
        Commands::Import => {
            let db_conn = &mut db::establish_connection()?;
            let redis_conn = &mut cache::establish_connection()?;

            let posts = ScanDir::dirs()
                .read("/blog", |iter| {
                    iter.map(|(entry, _)| {
                        let path = entry.path().join("post.md");
                        let src = fs::read_to_string(&path)?;
                        let frontmatter = match fronma::parser::parse::<PostFrontmatter>(&src) {
                            Ok(x) => x,
                            Err(x) => bail!("Error parsing frontmatter: {:?}", x),
                        };

                        Ok(Post {
                            path,
                            frontmatter: frontmatter.headers,
                            content: frontmatter.body.to_string(),
                        })
                    })
                    .collect::<Result<Vec<_>>>()
                })??
                .into_iter()
                .map(|post| -> Result<_> {
                    let trimmed = PostFrontmatter {
                        id: post.frontmatter.id,
                        name: post.frontmatter.name.trim().to_string(),
                        slug: post.frontmatter.slug.trim().to_string(),
                        description: post.frontmatter.description.trim().to_string(),
                        published_at: post.frontmatter.published_at,
                        edited_at: post.frontmatter.edited_at,
                        active: post.frontmatter.active,
                    };
                    let content = post.content.trim();

                    if let Some(id) = trimmed.id {
                        diesel::update(db::schema::posts::table)
                            .filter(db::schema::posts::id.eq(id))
                            .set(db::models::UpdatePost {
                                name: Some(&trimmed.name),
                                slug: Some(&trimmed.slug),
                                description: Some(&trimmed.description),
                                content: Some(content),
                                published_at: Some(trimmed.published_at),
                                edited_at: Some(trimmed.edited_at),
                                active: Some(trimmed.active),
                            })
                            .execute(db_conn)?;

                        Ok(id)
                    } else {
                        let id = if let Some(id) = db::schema::posts::table
                            .select(db::schema::posts::id)
                            .filter(db::schema::posts::slug.eq(&trimmed.slug))
                            .first::<i32>(db_conn)
                            .optional()?
                        {
                            diesel::update(db::schema::posts::table)
                                .filter(db::schema::posts::id.eq(id))
                                .set(db::models::UpdatePost {
                                    name: Some(&trimmed.name),
                                    slug: None,
                                    description: Some(&trimmed.description),
                                    content: Some(content),
                                    published_at: Some(trimmed.published_at),
                                    edited_at: Some(trimmed.edited_at),
                                    active: Some(trimmed.active),
                                })
                                .execute(db_conn)?;

                            id
                        } else {
                            diesel::insert_into(db::schema::posts::table)
                                .values(db::models::NewPost {
                                    name: &trimmed.name,
                                    slug: &trimmed.slug,
                                    description: &trimmed.description,
                                    content: content,
                                    published_at: trimmed.published_at,
                                    edited_at: trimmed.edited_at,
                                    active: trimmed.active,
                                })
                                .returning(db::schema::posts::id)
                                .get_result::<i32>(db_conn)?
                        };

                        fs::write(
                            post.path,
                            format!(
                                "---\n{}---\n\n{}\n",
                                serde_yaml::to_string(&PostFrontmatter {
                                    id: Some(id),
                                    ..trimmed
                                })?,
                                content
                            ),
                        )?;

                        Ok(id)
                    }
                })
                .collect::<Result<Vec<_>>>()?;

            let ids = db::schema::posts::table
                .select(db::schema::posts::id)
                .load::<i32>(db_conn)?;

            diesel::delete(
                db::schema::posts::table
                    .filter(diesel::dsl::not(db::schema::posts::id.eq_any(posts))),
            )
            .execute(db_conn)?;

            for id in ids {
                redis::cmd("DEL")
                    .arg(cache::keys::post_content(id))
                    .query::<()>(redis_conn)?;
            }

            Ok(())
        }
        Commands::Clear => {
            let db_conn = &mut db::establish_connection()?;
            let redis_conn = &mut cache::establish_connection()?;

            for id in db::schema::posts::table
                .select(db::schema::posts::id)
                .load::<i32>(db_conn)?
            {
                redis::cmd("DEL")
                    .arg(cache::keys::post_content(id))
                    .query::<()>(redis_conn)?;
            }

            Ok(())
        }
    }
}
